with Ada.Text_IO; use Ada.Text_IO;

package body SoE is

   --  each instance of process Sieve gets an own Id
   Assigned_Id : Natural := 0;

   --+-------------------------------------
   function New_Sieve (B : access Buffer_T)
     return Sieve_Ref is
   begin
      Assigned_Id := Assigned_Id + 1;
      --  this function returns a new process instance
      --+ of type Sieve_T
      return new Sieve_T (B, Assigned_Id);
   end New_Sieve;
   --+-------------------------------------

   --+-----------------------
   protected body Buffer_T is
      --  values can be written into A_Container
      --+ only if it is not full
      --+ hence Put is a potentially blocking operation
      entry Put (Item : Integer) when Contents < Limit is
      begin
         A_Container (Insert) := Item;
         Insert   := Insert + 1;
         Contents := Contents + 1;
      end Put;
      --+--------------------
      --  values can be fetched from A_Container
      --+ only if it is not empty
      --+ hence Get is a potentially blocking operation
      entry Get (Item : out Integer) when Contents > 0 is
      begin
         Item     := A_Container (Extract);
         Extract  := Extract + 1;
         Contents := Contents - 1;
      end Get;
      --+--------------------
   end Buffer_T;
   --+-----------------------

   --+-----------------------
   task body Odd is
      Num : Positive;
      Buf : aliased Buffer_T;
      --  process Odd has a child process S (instance of Sieve_T)
      --+ and gives it access to its own Buf object, which gives
      --+ protected access to A_Container
      S : Sieve_Ref := New_Sieve (Buf'Access);
   begin
      --  process S is now activated and starts running
      --+ concurrently to Odd
      Num := 3;
      while Num < Limit loop
         --  Odd fills A_Container with all odd numbers
         --+ from 3 until Limit, the first of which (3)
         --+ is known to be a prime
         Buf.Put (Num);
         --  Odd slows down to rest a little in between
         delay 0.2;
         Num := Num + 2;
      end loop;
      --  process Odd will complete when it will have passed down
      --+ the last odd number to its Sieve child, and will await the
      --+ termination of it to be able to terminate itself
   end Odd;
   --+-----------------------

   --+-----------------------
   task body Sieve_T is
      New_Buf    : aliased Buffer_T;
      Next_Sieve : Sieve_Ref;
      Prime, Num : Positive;
   begin
      --  Sieve fetches the first value from the array
      --+ it shares with its master via the protected buffer B,
      --+ which was passed to it as a discriminant
      --+ and this value is known to be a prime
      B.Get (Prime);
      loop
         --  Sieve tries the next odd number off B
         --+ until it finds one that is not divisible
         --+ by the one and only prime it knows
         B.Get (Num);
         if Num rem Prime /= 0 then
            exit;
         end if;
      end loop;
      --  As we proceed incrementally, Num is a new prime,
      --+ hence the Sieve process needs to pass Num down
      --+ to a child instance for it to continue the search
      --+ on all successive odd numbers not divisible by Prime
      --+ (which might be primes)
      Put_Line ("Sieve instance " & Natural'Image (Id) &
                " found prime number" & Natural'Image (Num));
      Next_Sieve := New_Sieve (New_Buf'Access);
      New_Buf.Put (Num);
      loop
         B.Get (Num);
         if Num rem Prime /= 0 then
            New_Buf.Put (Num);
         end if;
      end loop;
      --  what can we say about the termination of
      --+ the hierarchy of instance(s) of Sieve_T that
      --+ this code creates?
   end Sieve_T;
   --+-----------------------

end SoE;
